unit RVInsertMultilineText;

interface

uses
  SysUtils, Classes,
  RVItem, RVUni, RVTypes, RVStyle, RVStrFuncs, CRVData, RVInsertItems;

{
  Inserts multiline text Value in the position (RVData, ItemNo, Pos) using
  StyleNo text style.
  After inserting, (RVData, ItemNo, Pos) is a position after the inserted text.

  If StyleNo = -1, the procedure uses RVData.GetItemStyle(ItemNo).

  Value may contain tab characters, but cannot contain page break characters.
}
procedure InsertMultilineText(RVData: TCustomRVData;
  var ItemNo, Pos: Integer; const Value: TRVUnicodeString; StyleNo: Integer = -1);

implementation

procedure InsertMultilineText(RVData: TCustomRVData;
  var ItemNo, Pos: Integer; const Value: TRVUnicodeString; StyleNo: Integer);
var
  s1, s2, s3, s4: TRVUnicodeString;
  sl: TStringList;
  i, OldStyleNo, ParaNo: Integer;
  Tag: TRVTag;
  {............................................................................}
  function GetLine(sl: TStringList; Index: Integer): TRVUnicodeString;
  begin
    {$IFDEF UNICODE}
    Result := sl[Index];
    {$ELSE}
    Result := Utf8Decode(sl[Index]);
    {$ENDIF}
  end;
  {............................................................................}
  // If RVStyle.SpacesInTab > 0, replacing all tabs in sl to SpacesInTab spaces
  procedure ReplaceTabs;
  var s, Spaces: String;
      i, j, Count: Integer;
  begin
    Count := RVData.GetRVStyle.SpacesInTab;
    if Count = 0 then
      exit;
    SetLength(Spaces, Count);
    for i := 1 to Count do
      Spaces[i] := ' ';
    for i := 0 to sl.Count-1 do
    begin
      s := sl[i];
      for j := Length(s) downto 1 do
        if s[j] = #9 then
        begin
          Delete(s, j, 1);
          Insert(Spaces, s, j);
        end;
      sl[i] := s;
    end;
  end;
  {............................................................................}
  // Inserting tab in the ItemNo-th position. Uses ParaNo and StyleNo variables.
  procedure InsertTab(ItemNo: Integer; FromNewLine: Boolean);
  var
    tabitem: TRVTabItemInfo;
    s: TRVUnicodeString;
  begin
    tabitem := TRVTabItemInfo.Create(RVData);
    tabitem.StyleNo := rvsTab;
    tabitem.TextStyleNo := StyleNo;
    tabitem.ParaNo := ParaNo;
    if not FromNewLine then
      tabitem.SameAsPrev := True;
    s := '';
    tabitem.Inserting(RVData, s, False);
    RVData.Items.InsertObject(ItemNo, s, tabitem);
    tabitem.Inserted(RVData, ItemNo);
  end;
  {............................................................................}
  // Inserting s in the ItemNo-th position. s may contain tabs.
  // Adjusts ItemNo. Uses ParaNo and StyleNo variables.
  procedure InsertString(var ItemNo: Integer; s: TRVUnicodeString; FromNewLine: Boolean);
  var
    p, LParaNo: Integer;
    s2: TRVUnicodeString;
    Added: Boolean;
  begin
    Added := False;
    if not FromNewLine then
      LParaNo := -1
    else
      LParaNo := ParaNo;
    p := RVGet1Index(RVPosW(#9, s));
    while p > 0 do
    begin
      s2 := Copy(s, 1, p-1);
      if s2<>'' then
      begin
        RVInsertStringW(RVData, ItemNo, s2, StyleNo, LParaNo, Tag);
        LParaNo := -1;
        inc(ItemNo);
      end;
      InsertTab(ItemNo, LParaNo >= 0);
      inc(ItemNo);
      LParaNo := -1;
      s := Copy(s, p + 1, MaxInt);
      p := RVGet1Index(RVPosW(#9, s));
      Added := True;
    end;
    if (s<>'') or (not Added and FromNewLine) then
    begin
      RVInsertStringW(RVData, ItemNo, s, StyleNo, LParaNo, Tag);
      inc(ItemNo);
    end;
  end;
  {............................................................................}
begin
  if Value = '' then
    exit;
  OldStyleNo := RVData.GetItemStyle(ItemNo);
  if StyleNo < 0 then
    StyleNo := OldStyleNo;
  ParaNo := RVData.GetItemPara(ItemNo);
  Tag := RVData.GetItemTag(ItemNo);
  s1 := RVData.GetItemTextW(ItemNo);
  s2 := Copy(s1, Pos, Length(s1));
  s1 := Copy(s1, 1, Pos-1);
  sl := TStringList.Create;
  try
    sl.Text := {$IFDEF UNICODE}Value{$ELSE}Utf8Encode(Value){$ENDIF};
    ReplaceTabs;

    { workaround - TStringList ignores the last cr lf }
    if (Length(Value)>1) and
       (Value[Length(Value)-1]=#13) and (Value[Length(Value)]=#10) then
      sl.Add('');
    { end of workaround }

    i := System.Pos(#9, sl[0]);
    s3 := GetLine(sl, 0);
    s4 := '';
    if i > 0 then
    begin
      s4 := Copy(s3, i, MaxInt);
      s3 := Copy(s3, 1, i-1);    
    end;

    if StyleNo = OldStyleNo then
    begin
      if (sl.Count = 1) and (i = 0) then
      begin
        RVData.SetItemTextW(ItemNo, s1 + Value + s2);
        inc(Pos, Length(Value));
        exit;
      end;
      RVData.SetItemTextW(ItemNo, s1 + s3);
      inc(ItemNo);
      if sl.Count = 1 then
        s4 := s4 + s2;
      if s4 <> '' then
        InsertString(ItemNo, s4, False);
    end
    else
    begin
      if s1 = '' then
      begin
        RVData.GetItem(ItemNo).StyleNo := StyleNo;
        RVData.SetItemTextW(ItemNo, s3);
        Inc(ItemNo);
        if s4 <> '' then
          InsertString(ItemNo, s4, False);
      end
      else
      begin
        RVData.SetItemTextW(ItemNo, s1);
        inc(ItemNo);
        InsertString(ItemNo, GetLine(sl, 0), False);
      end;
    end;
    for i := 1 to sl.Count-1 do
    begin
      s1 := GetLine(sl, i);
      if (i = sl.Count-1) and (StyleNo = OldStyleNo) then
        s1 := s1 + s2;
      InsertString(ItemNo, s1, True);
    end;
    dec(ItemNo);
    Pos := RVData.GetOffsAfterItem(ItemNo);
    if StyleNo = OldStyleNo then
      dec(Pos, Length(s2));
    if (StyleNo <> OldStyleNo) and (s2 <> '') then
      RVInsertStringW(RVData, ItemNo + 1, s2, OldStyleNo, -1, Tag);
  finally
    sl.Free;
  end;
end;


end.
